package com.fastcashier.lib_common.net

import okhttp3.*
import okhttp3.internal.http.promisesBody
import okhttp3.internal.platform.Platform
import okhttp3.internal.platform.Platform.Companion.INFO
import okio.Buffer
import java.io.EOFException
import java.io.IOException
import java.nio.charset.Charset
import java.nio.charset.UnsupportedCharsetException
import java.util.concurrent.TimeUnit

/**
 * 基于原本的MiyaHttpLoggingInterceptor,本类做了如下改动：
 * 1. 请求整体写入，响应整体写入。但是请求和响应本身并不同时整体写入
 */
class MiyaHttpLoggingInterceptor @JvmOverloads constructor(private val logger: Logger = Logger.DEFAULT) :
    Interceptor {
    enum class Level {
        /**
         * No logs.
         */
        NONE,

        /**
         * Logs request and response lines.
         *
         *
         * Example:
         * <pre>`--> POST /greeting http/1.1 (3-byte body)
         *
         * <-- 200 OK (22ms, 6-byte body)
        `</pre> *
         */
        BASIC,

        /**
         * Logs request and response lines and their respective headers.
         *
         *
         * Example:
         * <pre>`--> POST /greeting http/1.1
         * Host: example.com
         * Content-Type: plain/text
         * Content-Length: 3
         * --> END POST
         *
         * <-- 200 OK (22ms)
         * Content-Type: plain/text
         * Content-Length: 6
         * <-- END HTTP
        `</pre> *
         */
        HEADERS,

        /**
         * Logs request and response lines and their respective headers and bodies (if present).
         *
         *
         * Example:
         * <pre>`--> POST /greeting http/1.1
         * Host: example.com
         * Content-Type: plain/text
         * Content-Length: 3
         *
         * Hi?
         * --> END POST
         *
         * <-- 200 OK (22ms)
         * Content-Type: plain/text
         * Content-Length: 6
         *
         * Hello!
         * <-- END HTTP
        `</pre> *
         */
        BODY
    }

    interface Logger {
        fun log(message: String?)

        companion object {
            /**
             * A [Logger] defaults output appropriate for the current platform.
             */
            val DEFAULT: Logger = object : Logger {
                override fun log(message: String?) {
                    Platform.get().log(message!!, INFO, null)
                }
            }
        }
    }

    @Volatile
    var level = Level.NONE

    /**
     * Change the level at which this interceptor logs.
     */
    fun setLevel(level: Level?): MiyaHttpLoggingInterceptor {
        if (level == null) throw NullPointerException("level == null. Use Level.NONE instead.")
        this.level = level
        return this
    }

    @Throws(IOException::class)
    override fun intercept(chain: Interceptor.Chain): Response {
        val level = level
        val request: Request = chain.request()
        if (level == Level.NONE) {
            return chain.proceed(request)
        }
        val logBody = level == Level.BODY
        val logHeaders = logBody || level == Level.HEADERS
        val requestBody = request.body
        val hasRequestBody = requestBody != null
        val connection: Connection? = chain.connection()
        val protocol = (connection?.protocol() ?: Protocol.HTTP_1_1)

        var requestStartMessage = "--> " + request.method + ' ' + request.url + ' ' + protocol
        if (!logHeaders && hasRequestBody) {
            requestStartMessage += " (" + requestBody!!.contentLength() + "-byte body)"
        }
        val sbRequest = StringBuilder()
        sbRequest.append(requestStartMessage)
        sbRequest.append("\n")
        if (logHeaders) {
            if (hasRequestBody) {
                // Request body headers are only present when installed as a network interceptor. Force
                // them to be included (when available) so there values are known.
                if (requestBody!!.contentType() != null) {
                    sbRequest.append("Content-Type: ").append(requestBody.contentType())
                        .append("\n")
                }
                if (requestBody.contentLength() != -1L) {
                    sbRequest.append("Content-Length: ").append(requestBody.contentLength())
                        .append("\n")
                }
            }
            val headers = request.headers
            var i = 0
            val count = headers.size
            while (i < count) {
                val name = headers.name(i)
                // Skip headers from the request body as they are explicitly logged above.
                if (!"Content-Type".equals(
                        name,
                        ignoreCase = true
                    ) && !"Content-Length".equals(name, ignoreCase = true)
                ) {
                    sbRequest.append(name).append(": ").append(headers.value(i)).append("\n")
                }
                i++
            }
            if (!logBody || !hasRequestBody) {
                sbRequest.append("--> END ").append(request.method).append("\n")
            } else if (bodyEncoded(request.headers)) {
                sbRequest.append("--> END ").append(request.method)
                    .append(" (encoded body omitted)").append("\n")
            } else {
                val buffer = Buffer()
                requestBody!!.writeTo(buffer)
                var charset = UTF8
                val contentType = requestBody.contentType()
                if (contentType != null) {
                    charset = contentType.charset(UTF8)
                }
                if (isPlaintext(buffer)) {
                    sbRequest.append(buffer.readString(charset!!)).append("\n")
                    sbRequest.append("--> END ").append(request.method).append(" (").append(
                        requestBody.contentLength()
                    ).append("-byte body)").append("\n")
                } else {
                    sbRequest.append("--> END ").append(request.method).append(" (binary ").append(
                        requestBody.contentLength()
                    ).append("-byte body omitted)").append("\n")
                }
            }
        }
        logger.log(sbRequest.toString()) //一次写入
        val sbResponse = StringBuilder()
        val startNs = System.nanoTime()
        val response: Response
        response = try {
            chain.proceed(request)
        } catch (e: Exception) {
            sbResponse.append("<-- HTTP FAILED: ").append(e).append("\n")
            logger.log(sbResponse.toString())
            throw e
        }
        val tookMs = TimeUnit.NANOSECONDS.toMillis(System.nanoTime() - startNs)
        val responseBody = response.body
        val contentLength = responseBody!!.contentLength()
        val bodySize = if (contentLength != -1L) "$contentLength-byte" else "unknown-length"
        sbResponse.append("<-- ")
            .append(response.code)
            .append(' ')
            .append(response.message)
            .append(' ')
            .append(response.request.url)
            .append(" (")
            .append(tookMs)
            .append("ms")
            .append(if (!logHeaders) ", $bodySize body" else "")
            .append(')')
            .append("\n")
        if (logHeaders) {
            val headers = response.headers
            var i = 0
            val count = headers.size
            while (i < count) {
                sbResponse.append(headers.name(i)).append(": ").append(headers.value(i))
                    .append("\n")
                i++
            }
            if (!logBody || !response.promisesBody()) {
                sbResponse.append("<-- END HTTP").append("\n")
            } else if (bodyEncoded(response.headers)) {
                sbResponse.append("<-- END HTTP (encoded body omitted)").append("\n")
            } else {
                val source = responseBody.source()
                source.request(Long.MAX_VALUE) // Buffer the entire body.
                val buffer = source.buffer()
                var charset = UTF8
                val contentType = responseBody.contentType()
                if (contentType != null) {
                    charset = try {
                        contentType.charset(UTF8)
                    } catch (e: UnsupportedCharsetException) {
                        sbResponse.append("Couldn't decode the response body; charset is likely malformed.")
                            .append("\n")
                        sbResponse.append("<-- END HTTP").append("\n")
                        logger.log(sbResponse.toString())
                        return response
                    }
                }
                if (!isPlaintext(buffer)) {
                    sbResponse.append("<-- END HTTP (binary ").append(buffer.size)
                        .append("-byte body omitted)").append("\n")
                    logger.log(sbResponse.toString())
                    return response
                }
                if (contentLength != 0L) {
                    sbResponse.append(buffer.clone().readString(charset!!)).append("\n")
                }
                sbResponse.append("<-- END HTTP (").append(buffer.size).append("-byte body)")
                    .append("\n")
            }
        }
        logger.log(sbResponse.toString())
        return response
    }

    private fun bodyEncoded(headers: Headers): Boolean {
        val contentEncoding = headers["Content-Encoding"]
        return contentEncoding != null && !contentEncoding.equals("identity", ignoreCase = true)
    }

    companion object {
        private val UTF8 = Charset.forName("UTF-8")

        /**
         * Returns true if the body in question probably contains human readable text. Uses a small sample
         * of code points to detect unicode control characters commonly used in binary file signatures.
         */
        fun isPlaintext(buffer: Buffer): Boolean {
            return try {
                val prefix = Buffer()
                val byteCount = if (buffer.size < 64) buffer.size else 64
                buffer.copyTo(prefix, 0, byteCount)
                for (i in 0..15) {
                    if (prefix.exhausted()) {
                        break
                    }
                    val codePoint = prefix.readUtf8CodePoint()
                    if (Character.isISOControl(codePoint) && !Character.isWhitespace(codePoint)) {
                        return false
                    }
                }
                true
            } catch (e: EOFException) {
                false // Truncated UTF-8 sequence.
            }
        }
    }
}